from typing import List
from UniversalParser._tools import patt_template_replace
from .templates import *

def parse_html_table(
        datas: List[List[List[str]]]
        , title: str = 'Table'
        , caption: str = ''
        , tfoot: str = ''
    ) -> str:

    tables = ''

    for rows in datas:

        tr_ths = rows[0]
        tr_tds = rows[1:]

        insert_tr_ths = ''
        insert_tr_tds = ''

        t_str = '<tr>'
        for th in tr_ths:
            t_str += f'<th>{th}</th>'
        t_str += '</tr>'
        insert_tr_ths += t_str

        for tr_td in tr_tds:
            t_str = '<tr>'
            for td in tr_td:
                t_str += f'<td>{td}</td>'
            t_str += '</tr>'
            insert_tr_tds += t_str

        tables += patt_template_replace(
            TABLE
            , tr_ths = insert_tr_ths
            , tr_tds = insert_tr_tds
        )
    
    return patt_template_replace(
        HTML
        , title = title
        , tables = tables
    )



# try:
#     import yaml
# except:
#     raise ImportError('Please `pip install -r requirements.txt`.')
# else:
#     versions = []
#     versions.extend([
#         yaml.__version__ == '5.4.1'
#     ])

#     if not all(versions):
#         raise ImportError('''\
# requirements: yaml==5.4.1\
# ''')



HTML_SINGLE_TAGS = [
    'link',
    'meta',
    'br',
    'area',
    'base',
    'col',
    'command',
    'embed',
    'hr',
    'img',
    'input',
    'keygen',
    'param',
    'source',
    'track',
    'wbr',
]


from typing import Dict, List
from functools import reduce
from zipfile import ZipFile
from ._exception import *
import re

def __intersection(obj1, obj2):
    result = []
    for _1 in obj1:
        for _2 in obj2:
            if _1 == _2:
                result.append(_1)
                break
    return result

def find_intersection(datas: List[List[int]]) -> List[int]:
    return list(reduce(__intersection, datas))

def patt_template_replace(data: str, *args, patt=None, **kwargs):
    if patt is None:
        patt = re.compile(r'\$\{(.+?)\}')
    def deal(x):
        if x.group(1) in kwargs:
            return kwargs[x.group(1)]
        else:
            return x.group(0)
    if patt.search(data):
        return patt.sub(deal, data)
    else:
        return data

def read_content_from_zip(
        zip_file_path: str
        , read_file_path: str='word/document.xml'
        , encoding: str='utf-8'
    ) -> str:
    try:
        with ZipFile(zip_file_path) as zip_obj:
            with zip_obj.open(read_file_path) as op_file:
                content = op_file.read().decode(encoding)
    except Exception as e:
        raise ZipParseError(f'zip-file can not be parsed.{e}')

    return content

def read_content_from_zipobj(
        zip_obj
        , read_file_path: str
        , encoding: str='utf-8'
    ):
    try:
        with zip_obj.open(read_file_path) as op_file:
            content = op_file.read().decode(encoding)
    except Exception as e:
        raise ZipParseError(f'zip-file can not be parsed.{e}')

    return content

def save_docs_with_modify_by_move(
        zip_obj: ZipFile
        , modify_files: Dict[str, str] # {'word/document.xml': '...', ...}
        , save_zip_path: str = 'output.docx'
        # , encoding: str = 'utf-8'
    ):
    out_zip_obj = ZipFile(save_zip_path, 'w')
    for item in zip_obj.infolist():
        if item.filename in modify_files:
            out_zip_obj.writestr(item, modify_files[item.filename])
        else:
            buffer = zip_obj.read(item.filename)
            out_zip_obj.writestr(item, buffer)
    out_zip_obj.close()



from typing import TypeVar

# table
ROW = TypeVar('ROW')
COL = TypeVar('COL')

class MicroTagType:
    WORD = 'word'
    EXCEL = 'excel'
    PPT = 'ppt'

class WordTag:
    TABLE = 'w:tbl'
    TR = 'w:tr'
    TC = 'w:tc'
    P = 'w:p'
    R = 'w:r'
    TEXT = 'w:t'
    TCPR = 'w:tcPr'
    RPR = 'w:rPr'
    RFONTS = 'w:rFonts'
    VERTALIGN = 'w:vertAlign'
    LANG = 'w:lang'

    class Attribute:
        GRIDSPAN = 'w:gridSpan' # w:val="2" [combine cols]
        VMERGE = 'w:vMerge' # w:val="restart"  w:val="continue" [combine rows]
        VAL = 'w:val'
        HINT = 'w:hint'
        EASTASIA = 'w:eastAsia'

class PowerPointTag:
    TABLE = 'a:tbl'
    TR = 'a:tr'
    TC = 'a:tc'
    P = 'a:p'
    R = 'a:r'
    TEXT = 'a:t'
    
    class Attribute:
        pass

class ExcelTag:
    TABLE = 'sheetData'
    TR = 'row'
    TC = 'c'
    TEXT = 'v'

    class Attribute:
        LOC = 'r'
        SPANS = 'spans'



from collections import OrderedDict
from typing import Dict, List, Optional, Tuple
from xml.parsers.expat import ParserCreate, XMLParserType
from UniversalParser._ntype import *

__author__ = 'jiyang'
__version__ = '0.0.1'   
__license__ = 'MIT'

def parse_odict(xml_data: str
        , encoding: str = None
        , namespace_separator: str = None
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , combine_cdata: bool = True
        , include_comment: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , loc_key: str = LOC_KEY
        , include_loc: bool = False
    ) -> OrderedDict:

    xml_declares = {'version': '1.0', 'encoding': f'{encoding}', 'standalone': -1}
    documentation: OrderedDict = OrderedDict({})
    stack_nodes: List[OrderedDict] = [[documentation, [], 0], ]
    is_cdata: bool = False
    temp_loc: int = 0
    flag_out: int = 0

    def onXmlDeclHandler(version: int, encoding: Optional[str], standalone: int):
        nonlocal xml_declares
        xml_declares['version'] = version
        xml_declares['encoding'] = encoding
        xml_declares['standalone'] = standalone

    def onStartElementHandler(name: str, attrs: List[str]):
        nonlocal stack_nodes, temp_loc, flag_out
        if 0 == flag_out: # only in.
            temp_loc = 0
        current_node = stack_nodes[-1][0]
        node = OrderedDict({})
        stack_nodes.append([node, [], temp_loc])
        for i in range(len(attrs)//2):
            attr_name, attr_value = attrs[i*2], attrs[i*2+1]
            node[f'{attr_prefix}{attr_name}'] = attr_value
        if include_loc:
            node[f'{attr_prefix}{loc_key}'] = f'{temp_loc}'
        if name in current_node:
            sub_node = current_node[name]
            if isinstance(sub_node, OrderedDict):
                current_node[name] = [sub_node, node, ]
            elif isinstance(sub_node, list):
                current_node[name].append(node)
        else:
            current_node.update(OrderedDict({name: node}))
        flag_out = 0

    def onEndElementHandler(name: str):
        nonlocal stack_nodes, temp_loc, flag_out
        node = stack_nodes.pop()
        node[0][cdata_key] = cdata_separator.join(node[1])

        flag_out = 1
        temp_loc = f'{int(node[2])+1}'

    def onStartCdataSectionHandler():
        nonlocal is_cdata
        is_cdata = True

    def onEndCdataSectionHandler():
        nonlocal is_cdata
        is_cdata = False

    def onCommentHandler(data: str):
        nonlocal stack_nodes
        if include_comment:
            node = stack_nodes[-1][0]
            if comment_key in node:
                node[comment_key].append(data.strip())
            else:
                node[comment_key] = [data.strip(), ]

    def onCharacterDataHandler(data: str):
        nonlocal stack_nodes, is_cdata
        t_data = data.strip()
        if not combine_cdata and is_cdata:
            node = stack_nodes[-1][0]
            if cdata_self_key not in node:
                node[cdata_self_key] = [t_data, ]
            else:
                node[cdata_self_key].append(t_data)
        else:
            if t_data:
                stack_nodes[-1][1].append(t_data)

    xmlParser: XMLParserType = ParserCreate(encoding=encoding, namespace_separator=namespace_separator)
    xmlParser.XmlDeclHandler = onXmlDeclHandler
    xmlParser.StartElementHandler = onStartElementHandler
    xmlParser.EndElementHandler = onEndElementHandler
    xmlParser.CommentHandler = onCommentHandler
    xmlParser.CharacterDataHandler = onCharacterDataHandler
    xmlParser.StartCdataSectionHandler = onStartCdataSectionHandler
    xmlParser.EndCdataSectionHandler = onEndCdataSectionHandler

    xmlParser.buffer_text = True
    xmlParser.ordered_attributes = True

    xmlParser.Parse(xml_data, 1)

    return documentation, xml_declares



from typing import TypeVar, Union

class TextType:
    STR = 0
    INT = 1
    FLOAT = 2
    DECIMAL = 3

class ParserType:
    XML = 0
    JSON = 1
    DICT = 2
    YAML = 3
    HTML = 4

class AnalysisMode:
    RECURSION_OLD = 0
    RECOMMEND = 1
    DIRECT = 2

class SM:
    text = 0
    attr_names = 1
    tag = 2
    ancestor = 3
    descendants = 4
    siblings = 5
    children = 6
    parent = 7

Type_Path = TypeVar('Type_Path')
Type_JSON = Union[Type_Path, str]

# analysis area.
ATTR_PREFIX = '@'
CDATA_KEY = '#text'
REAL_CDATA_KEY = 'text_'
CDATA_SELF_KEY = '#CDATA'
COMMENT_KEY = '#comment'
LOC_KEY = '__loc__'
CDATA_SEPARATOR = ' '



__all__ = [
    'parse_tables',
]

from .BaseTable import *
from UniversalParser._ntype import *
from UniversalParser.micro._ntype import *

class ExcelTable(BaseTable):
    
    def __init__(self
            , zip_file_path: str
            , read_file_path: str = 'ppt/slides/slide1.xml'
            , encoding: str = 'utf-8'
            , mode: str = 'a' # do not use `w` mode, remember!
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , real_cdata_key: str = REAL_CDATA_KEY
            , cdata_self_key: str = CDATA_SELF_KEY
            , comment_key: str = COMMENT_KEY
            , analysis_text: bool = False
            , cdata_separator: str = CDATA_SEPARATOR
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
        ) -> None:

        super().__init__(zip_file_path
            , MicroTagType.PPT
            , read_file_path = read_file_path
            , encoding = encoding
            , mode = mode
            , analysis_text = analysis_text
            , attr_prefix = attr_prefix
            , cdata_key = cdata_key
            , real_cdata_key = real_cdata_key
            , cdata_self_key = cdata_self_key
            , comment_key = comment_key
            , cdata_separator = cdata_separator
            , analysis_mode = analysis_mode
        )

    def refresh_tables(self): pass

    def modify_cell_by_coordinate(self): pass

    def batch_modify_cells_by_coordinate(self): pass

def parse_tables(
        zip_file_path: str
        , encoding: str = 'utf-8'
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , analysis_text: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
    ) -> ExcelTable:
    read_file_path: str = f'ppt/slides/slide1.xml'
    return ExcelTable(
        zip_file_path = zip_file_path
        , read_file_path = read_file_path
        , encoding = encoding
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , analysis_text = analysis_text
        , cdata_separator = cdata_separator
        , analysis_mode = analysis_mode
    )



from ._rely import *
from ._ntype import *
from ._exception import *
from ._collections import *
from collections import deque
from typing import Any, Callable, Dict, List, OrderedDict as Type_OrderedDict, Tuple, Union
# import warnings

class ChainXML:

    xml = None

    def __init__(self
            , doc: Union[Type_OrderedDict, dict]
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , real_cdata_key: str = REAL_CDATA_KEY
            , cdata_self_key: str = CDATA_SELF_KEY
            , comment_key: str = COMMENT_KEY
            , cdata_separator: str = CDATA_SEPARATOR
            , loc_key: str = LOC_KEY
            , data_switch: int = ParserType.XML
            , analysis_text: bool = True
            , open_cdata: bool = True
            , open_comment: bool = True
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
        ) -> None:

        if data_switch in [ParserType.XML, ParserType.HTML, ]:
            if not isinstance(doc, OrderedDict) or 1 != len(doc):
                raise DocTypeError('`doc` type error. Only be one root.')
        elif data_switch in [ParserType.DICT, ParserType.YAML, ParserType.JSON, ]:
            if not isinstance(doc, dict):
                raise DocTypeError('`doc` type error.')
            if len(doc) < 0:
                raise DocTypeError('`doc` type error. Must be greater than or equal to one root.')
            elif 1 != len(doc):
                doc = {'root': doc}
                # warnings.warn('`doc` is greater than one root, the results may not be satisfactory. Auto add `root` node.')
        else:
            raise DocTypeError('Unknown `doc` type.')

        self.doc = doc
        self.attr_prefix = attr_prefix
        self.cdata_key = cdata_key
        self.real_cdata_key = real_cdata_key
        self.cdata_self_key = cdata_self_key
        self.comment_key = comment_key
        self.loc_key = loc_key
        self.cdata_separator = cdata_separator
        self.loc_attr_name = f'{attr_prefix}{loc_key}'
        ChainDict.cdata_key = self.real_cdata_key
        self.analysis_text = analysis_text
        self.open_cdata = open_cdata
        self.open_comment = open_comment

        self._attr_searcher = {}
        self._reverse_attr_name = {}
        self._id_nodes = {}
        self._id_tags = {}
        self._tag_ids = {}
        self._text_ids = {}
        self._textids = []
        self._contrast_ids = {}
        self._comment_ids = {}
        self._cdata_ids = {}
        self.xml = ChainDict(self) # ori xml-dict

        if AnalysisMode.RECURSION_OLD == analysis_mode:
            self._build_chain_by_recursion(self.doc, self.xml)
        elif AnalysisMode.RECOMMEND == analysis_mode:
            self._build_chain_recommend(self.doc, self.xml) # Extra support CDATA(not the same as text.) and comment.
        elif AnalysisMode.DIRECT == analysis_mode:
            self._build_chain_direct(self.doc, self.xml) # future.
        else:
            raise AnalysisTypeError('Analysis type error. Please choice one of exist types.')

    def registry_id_nodes(self, node: Any) -> None:
        self._id_nodes[id(node)] = node

    def remove_id_nodes(self, node: Any) -> None:
        if id(node) in self._id_nodes:
            self._id_nodes.pop(id(node))
        else:
            raise Exception('Node has been removed.')

    def registry_contrast_ids(self, sub_node: Any, node: ChainDict) -> None:
        self._contrast_ids[id(sub_node)] = id(node)

    def remove_contrast_ids(self, sub_node: Any) -> None:
        if id(sub_node) in self._contrast_ids:
            self._contrast_ids.pop(id(sub_node))
        else:
            raise Exception('Property has been removed.')

    def modify_contrast_ids(self, sub_node: Any, node: ChainDict) -> None:
        if id(sub_node) in self._contrast_ids:
            self._contrast_ids[id(sub_node)] = id(node)
        else:
            raise Exception('Property does not exist.')

    def registry_text_ids(self, text: str, node: Any) -> None:
        if self.analysis_text:
            if text in self._text_ids:
                if id(node) not in self._text_ids[text]:
                    self._text_ids[text].append(id(node))
                    self._textids.append(id(node))
                else:
                    raise Exception('Fatal error in parser! Do not re register.')
            else:
                self._text_ids[text] = [id(node), ]
                self._textids.append(id(node))
        else:
            if id(node) not in self._textids:
                self._textids.append(id(node))

    def registry_comment_ids(self, comments: List[str], node: Any) -> None:
        if self.open_comment:
            for comment in comments:
                if comment in self._comment_ids:
                    if id(node) not in self._comment_ids[comment]:
                        self._comment_ids[comment].append(id(node))
                    # else: # distinct
                    #     raise Exception('Fatal error in parser! Do not re register.')
                else:
                    self._comment_ids[comment] = [id(node), ]

    def registry_cdata_ids(self, cdatas: List[str], node: Any) -> None:
        if self.open_cdata:
            for cdata in cdatas:
                if cdata in self._cdata_ids:
                    if id(node) not in self._cdata_ids[cdata]:
                        self._cdata_ids[cdata].append(id(node))
                    # else: # distinct
                    #     raise Exception('Fatal error in parser! Do not re register.')
                else:
                    self._cdata_ids[cdata] = [id(node), ]

    def remove_text_ids(self, text: str, node: Any) -> None:
        if self.analysis_text:
            if text in self._text_ids:

                if id(node) in self._text_ids[text]:
                    self._text_ids[text].remove(id(node))
                    self._textids.remove(id(node))

                    if 0 == len(self._text_ids[text]):
                        self._text_ids.pop(text)
                else:
                    raise Exception('Property has been removed.')
            else:
                raise Exception('Property has been removed.')
        else:
            if id(node) in self._textids:
                self._textids.remove(id(node))

    def modify_text_ids(self, old_text: str, new_text: str, node: Any) -> None:
        self.remove_text_ids(old_text, node)
        self.registry_text_ids(new_text, node)

    def registry_id_tags(self, node: Any, tag: str) -> None:
        self._id_tags[id(node)] = tag

    def remove_id_tags(self, node: Any, tag: str) -> None:
        if id(node) in self._id_tags:
            self._id_tags.pop(id(node))
        else:
            raise Exception('Tag has been removed.')
    
    def modify_id_tags(self, node: Any, old_tag: str, new_tag: str) -> None:
        self.remove_id_tags(node, old_tag)
        self.registry_id_tags(node, new_tag)

    def registry_tag_ids(self, tag: str, node: Any) -> None:
        if tag in self._tag_ids:
            if id(node) not in self._tag_ids[tag]:
                self._tag_ids[tag].append(id(node))
            else:
                raise Exception('Fatal error in parser! Do not re register.')
        else:
            self._tag_ids[tag] = [id(node), ]

    def remove_tag_ids(self, tag: str, node: Any) -> None:
        if tag in self._tag_ids:
            if id(node) in self._tag_ids[tag]:
                self._tag_ids[tag].remove(id(node))

                if 0 == len(self._tag_ids[tag]):
                    self._tag_ids.pop(tag)
            else:
                raise Exception('Tag has been removed.')
        else:
            raise Exception('Tag has been removed.')
    
    def modify_tag_ids(self, tag: str, old_node: Any, new_node: Any) -> None:
        self.remove_tag_ids(tag, old_node)
        self.registry_tag_ids(tag, new_node)

    def registry_attr_searcher(self, search_key: str, node: Any) -> None:
        if search_key in self._attr_searcher:
            if id(node) not in self._attr_searcher[search_key]['nodes']:
                self._attr_searcher[search_key]['nodes'].append(id(node))
                self._attr_searcher[search_key]['count'] += 1
            else:
                raise Exception('Fatal error in parser! Do not re register.')
        else:
            self._attr_searcher[search_key] = {
                'count': 1,
                'nodes': [id(node), ]
            }

    def remove_attr_searcher(self, search_key: str, node: Any) -> None:
        if search_key in self._attr_searcher:
            if id(node) in self._attr_searcher[search_key]['nodes']:
                self._attr_searcher[search_key]['nodes'].remove(id(node))
                self._attr_searcher[search_key]['count'] -= 1
                if 0 == self._attr_searcher[search_key]['count']:
                    self._attr_searcher.pop(search_key)
            else:
                raise Exception('Property has been removed.')
        else:
            raise Exception('Property has been removed.')

    def modify_attr_searcher(self, old_search_key: str, new_search_key: str, node: Any) -> None:
        self.remove_attr_searcher(old_search_key, node)
        self.registry_attr_searcher(new_search_key, node)

    def registry_reverse_attr_name(self, new_old: Tuple[str, str], node: Any) -> None:
        if new_old not in self._reverse_attr_name:
            self._reverse_attr_name[new_old] = [id(node), ]
        else:
            if id(node) not in self._reverse_attr_name[new_old]:
                self._reverse_attr_name[new_old].append(id(node))
            else:
                raise Exception('Fatal error in parser! Do not re register.')

    def remove_reverse_attr_name(self, new_old: Tuple[str, str], node: Any) -> None:
        if new_old in self._reverse_attr_name:
            if id(node) in self._reverse_attr_name[new_old]:
                self._reverse_attr_name[new_old].remove(id(node))
                if 0 == len(self._reverse_attr_name[new_old]):
                    self._reverse_attr_name.pop(new_old)
            else:
                raise Exception('Property has been removed.')
        else:
            raise Exception('Property has been removed.')

    def modify_reverse_attr_name(self, o_new_old: Tuple[str, str], n_new_old: str, node: Any) -> None:
        self.remove_reverse_attr_name(o_new_old, node)
        self.registry_reverse_attr_name(n_new_old, node)

    def _get_searchkeys_by_nodeid(self, node_id: int) -> str:
        search_keys = []
        for search_key, obj in self._attr_searcher.items():
            if node_id in obj['nodes']:
                search_keys.append(search_key)
        return search_keys

    def _get_newolds_by_nodeid(self, node_id: int) -> Tuple[str, str]:
        new_olds = []
        for new_old, node_ids in self._reverse_attr_name.items():
            if node_id in node_ids:
                new_olds.append(new_old)
        return new_olds

    def signal_del_node_base(self, node: Any) -> None:
        stack = [id(node), *self.descendants(id(node))]
        while stack:
            pop_node_id = stack.pop()
            pop_node = self._id_nodes[pop_node_id]
            parent_node = self._id_nodes[self.parent(pop_node_id)]

            search_keys = self._get_searchkeys_by_nodeid(pop_node_id)
            tag = self.get_tag(pop_node)
            new_olds = self._get_newolds_by_nodeid(pop_node_id)

            if isinstance(pop_node, ChainDict):
                text = pop_node[self.real_cdata_key]
                self.remove_text_ids(text, pop_node)

            if isinstance(parent_node, list):
                parent_node.remove(pop_node)
            else:
                if isinstance(parent_node[tag], ChainDict):
                    parent_node.pop(tag)
                elif isinstance(parent_node[tag], list):
                    parent_node[tag].remove(pop_node)
                else:
                    raise PopError("Pop node error, obj must be `ChainDict` type, and obj's parent node-type must be `ChainDict` or `list`.")

            for search_key in search_keys:
                self.remove_attr_searcher(search_key, pop_node)
            for new_old in new_olds:
                self.remove_reverse_attr_name(new_old, pop_node)
            self.remove_id_nodes(pop_node)
            self.remove_id_tags(pop_node, tag)
            self.remove_tag_ids(tag, pop_node)
            self.remove_contrast_ids(pop_node)
        
    def signal_add_node(self, sub_node: Any, node: Any, tag: str=None, text: str=None) -> None:
        self.registry_id_nodes(sub_node)
        self.registry_id_tags(sub_node, tag)
        self.registry_contrast_ids(sub_node, node)
        if tag is not None:
            self.registry_tag_ids(tag, sub_node)
        if text is not None:
            self.registry_text_ids(text, sub_node)

    def signal_move_node(self, move_node, from_node, to_node) -> None:
        if id(from_node) != self._contrast_ids[id(move_node)]:
            raise NoNodesFound('`from_node` error.')
        if id(from_node) == id(to_node):
            raise MoveError("`from_node` can't be the same as `to_node`.")
        self._contrast_ids[id(move_node)] = id(to_node)

    def signal_add_attr_base(self, node: Any, attr_name: str, attr_value: str) -> None:
        new_old = (attr_name, f'{self.attr_prefix}{attr_name}')
        search_key = f'{attr_name}={attr_value}'
        self.registry_reverse_attr_name(new_old, node)
        self.registry_attr_searcher(search_key, node)

    def signal_add_attr(self, node: Any, new_old: Tuple[str, str], search_key: str) -> None:
        self.registry_reverse_attr_name(new_old, node)
        self.registry_attr_searcher(search_key, node)

    def signal_modify_attr(self, old_search_key: str, new_search_key: str, node: Any):
        self.modify_attr_searcher(old_search_key, new_search_key, node)

    def signal_del_attr_base(self, node: Any, attr_name: str, attr_value: str) -> None:
        new_old = (attr_name, f'{self.attr_prefix}{attr_name}')
        search_key = f'{attr_name}={attr_value}'
        self.remove_reverse_attr_name(new_old, node)
        self.remove_attr_searcher(search_key, node)

    def signal_modify_text(self, old_text: str, new_text: str, node: Any) -> None:
        self.modify_text_ids(old_text, new_text, node)

    def _get_last_loc(self, node: ChainDict) -> str:
        max_loc = -1
        for child in self.get_children(node):
            if self.loc_key in child:
                if max_loc < int(child[self.loc_key]):
                    max_loc = int(child[self.loc_key])
        if max_loc != -1:
            return str(max_loc+1)
        else:
            return '0'

    def _build_chain_by_recursion(self, sub_node_v: Union[str, list, dict], parent_node: ChainDict = None, tag: str=None, flag: bool=False) -> None:

        self.registry_id_nodes(parent_node)

        if tag is None:
            if isinstance(sub_node_v, dict):
                for _tag, _sub_node_v in sub_node_v.items():
                    self._build_chain_by_recursion(_sub_node_v, parent_node, _tag)
                return
            else:
                raise DocTypeError('The XML document is malformed.')

        if tag.startswith(self.attr_prefix):
            new_old = (tag[len(self.attr_prefix):], tag)
            search_key = f'{new_old[0]}={sub_node_v}'
            parent_node[new_old[0]] = sub_node_v
            self.signal_add_attr(parent_node, new_old, search_key)
            
        elif self.cdata_key == tag:
            parent_node[self.real_cdata_key] = sub_node_v
            self.registry_text_ids(sub_node_v, parent_node)
            
        else:
            if sub_node_v is None:
                if not flag:
                    insert_node = ChainDict(self)
                    insert_node[self.real_cdata_key] = ''
                    parent_node[tag] = insert_node
                    self.signal_add_node(insert_node, parent_node, tag, '')

            elif isinstance(sub_node_v, dict):
                temp_node = ChainDict(self)
                
                if not flag:
                    parent_node[tag] = temp_node

                for _tag, _sub_node_v in sub_node_v.items():
                    if flag:
                        self._build_chain_by_recursion(_sub_node_v, parent_node, _tag)
                    else:
                        self._build_chain_by_recursion(_sub_node_v, temp_node, _tag)
                
                if not flag:
                    if self.real_cdata_key not in temp_node:
                        temp_node[self.real_cdata_key] = ''
                        self.signal_add_node(temp_node, parent_node, tag=tag, text='')
                    else:
                        self.signal_add_node(temp_node, parent_node, tag=tag)

            elif isinstance(sub_node_v, (list, tuple)):
                temp_list = []
                self.signal_add_node(temp_list, parent_node, tag=tag)

                for sub_node in sub_node_v:
                    temp_node = ChainDict(self)
                    temp_list.append(temp_node)

                    self._build_chain_by_recursion(sub_node, temp_node, tag, True)

                    if self.real_cdata_key not in temp_node:
                        temp_node[self.real_cdata_key] = ''
                        self.signal_add_node(temp_node, temp_list, tag, '')
                    else:
                        self.signal_add_node(temp_node, temp_list, tag)
                parent_node[tag] = temp_list
                
            else:
                if flag:
                    parent_node[self.real_cdata_key] = sub_node_v
                    self.registry_text_ids(sub_node_v, parent_node)
                else:
                    temp_node = ChainDict(self)
                    temp_node[self.real_cdata_key] = sub_node_v
                    self.signal_add_node(temp_node, parent_node, tag, sub_node_v)
                    parent_node[tag] = temp_node

    def _build_chain_recommend(self, datas: Union[dict, list], root_node: ChainDict) -> None:
        root_data = list(datas.items())[0]
        queue = deque([[root_data[0], root_data[1], root_node]])
        while queue:
            pop_obj = queue.popleft()
            pop_tag = pop_obj[0]
            pop_node = pop_obj[1]
            parent_node = pop_obj[2]
            self.registry_id_nodes(parent_node) # Only effect to first node.
            sub_nodes = []
            temp_node = ChainDict(self)

            for tag, value in pop_node.items():
                if tag.startswith(self.attr_prefix):
                    attr_name = tag[len(self.attr_prefix):]
                    temp_node[attr_name] = value
                    self.signal_add_attr_base(temp_node, attr_name, value)
                elif self.cdata_key == tag:
                    temp_node[self.real_cdata_key] = value
                    self.registry_text_ids(value, temp_node)
                elif self.cdata_self_key == tag:
                    temp_node[self.cdata_self_key] = value
                    self.registry_cdata_ids(value, temp_node)
                elif self.comment_key == tag:
                    temp_node[self.comment_key] = value # value: List
                    self.registry_comment_ids(value, temp_node)
                else:
                    if isinstance(value, dict):
                        sub_nodes.append([tag, value, temp_node])
                    elif isinstance(value, (list, tuple)):
                        sub_nodes.extend([[tag, _1, temp_node] for _1 in value])
                    else:
                        raise TypeError('Unknown status')

            if pop_tag not in parent_node:
                parent_node[pop_tag] = temp_node
                self.signal_add_node(temp_node, parent_node, tag=pop_tag)
            else:
                if isinstance(parent_node[pop_tag], ChainDict):
                    t_node = parent_node[pop_tag]
                    temp_list = []
                    parent_node[pop_tag] = temp_list
                    temp_list.append(t_node)
                    self.signal_add_node(temp_list, parent_node, tag=pop_tag)
                    self.signal_move_node(t_node, from_node=parent_node, to_node=temp_list)
                    temp_list.append(temp_node)
                    self.signal_add_node(temp_node, temp_list, tag=pop_tag)
                elif isinstance(parent_node[pop_tag], list):
                    parent_node[pop_tag].append(temp_node)
                    self.signal_add_node(temp_node, parent_node, tag=pop_tag)
                else:
                    raise TypeError('Unknown type.')
            
            queue.extendleft(reversed(sub_nodes))

    def _build_chain_direct(self, datas: dict, root: ChainDict) -> None:
        return

    def _build_by_copy(self, datas: Union[dict, list], root_node: ChainDict) -> None:
        queue = deque([])
        if isinstance(datas, dict):
            for _tag, _node in datas.items():
                if isinstance(_node, dict):
                    queue.append([_tag, _node, root_node])
                elif isinstance(_node, list):
                    queue.extendleft([[_tag, _1, root_node] for _1 in reversed(_node)])
                else:
                    raise TypeError('Only support `dict`.')
        else:
            raise TypeError('Only support `dict`.')

        while queue:
            pop_obj = queue.popleft()
            pop_tag = pop_obj[0]
            pop_node = pop_obj[1]
            parent_node = pop_obj[2]
            self.registry_id_nodes(parent_node) # Only effect to first node.
            sub_nodes = []
            temp_node = ChainDict(self)

            for tag, value in pop_node.items():
                if tag.startswith(self.attr_prefix):
                    if tag == self.loc_attr_name:
                        loc_value = self._get_last_loc(parent_node)
                        temp_node[self.loc_key] = loc_value
                        self.signal_add_attr_base(temp_node, self.loc_key, loc_value)
                    else:
                        attr_name = tag[len(self.attr_prefix):]
                        temp_node[attr_name] = value
                        self.signal_add_attr_base(temp_node, attr_name, value)
                elif self.cdata_key == tag:
                    temp_node[self.real_cdata_key] = value
                    self.registry_text_ids(value, temp_node)
                elif self.cdata_self_key == tag:
                    temp_node[self.cdata_self_key] = value
                    self.registry_cdata_ids(value, temp_node)
                elif self.comment_key == tag:
                    temp_node[self.comment_key] = value
                    self.registry_comment_ids(value, temp_node)
                else:
                    if isinstance(value, dict):
                        sub_nodes.append([tag, value, temp_node])
                    elif isinstance(value, (list, tuple)):
                        sub_nodes.extend([[tag, _1, temp_node] for _1 in value])
                    else:
                        raise TypeError('Unknown status')

            if pop_tag not in parent_node:
                parent_node[pop_tag] = temp_node
                self.signal_add_node(temp_node, parent_node, tag=pop_tag)
            else:
                if isinstance(parent_node[pop_tag], ChainDict):
                    t_node = parent_node[pop_tag]
                    temp_list = []
                    parent_node[pop_tag] = temp_list
                    temp_list.append(t_node)
                    self.signal_add_node(temp_list, parent_node, tag=tag)
                    self.signal_move_node(t_node, from_node=parent_node, to_node=temp_list)
                    temp_list.append(temp_node)
                    self.signal_add_node(temp_node, temp_list, tag=pop_tag)
                elif isinstance(parent_node[pop_tag], list):
                    parent_node[pop_tag].append(temp_node)
                    self.signal_add_node(temp_node, parent_node, tag=pop_tag)
                else:
                    raise TypeError('Unknown type.')
            
            queue.extendleft(reversed(sub_nodes))

    def exists(self, node_id: Union[int, ChainDict, list]) -> bool:
        if isinstance(node_id, (ChainDict, list)):
            node_id = id(node_id)
        if node_id in list(self._contrast_ids.keys()) + list(self._contrast_ids.values()):
            return True
        else:
            return False

    def is_parent_exist(self, node_id: Union[int, ChainDict, list]) -> bool:
        if isinstance(node_id, (ChainDict, list)):
            node_id = id(node_id)
        if node_id in self._contrast_ids:
            return True
        else:
            return False

    def parent(self, node_id: int) -> int:
        if self.is_parent_exist(node_id):
            return self._contrast_ids[node_id]
        else:
            if node_id in self._contrast_ids.values():
                return node_id
            else:
                raise NoNodesFound('`node_id` error.')

    def get_parent(self, node: ChainDict) -> ChainDict:
        return self._id_nodes[self.parent(id(node))]

    def children(self, p_id: int) -> List[int]:
        children_node_ids = []
        for t_c, t_p in self._contrast_ids.items():
            if t_p == p_id:
                children_node_ids.append(t_c)
        return children_node_ids

    def get_children(self, node: ChainDict, combine=True) -> List[ChainDict]:
        temp_list = [self._id_nodes[_1] for _1 in self.children(id(node))]
        if combine:
            result = []
            for obj in temp_list:
                if isinstance(obj, list):
                    result.extend(obj)
                else:
                    result.append(obj)
            return result
        else:
            return temp_list
    
    def siblings(self, node_id: int) -> List[int]:
        p_id = self.parent(node_id)
        siblings_node_ids = self.children(p_id)
        siblings_node_ids.remove(node_id)
        return siblings_node_ids

    def get_siblings(self, node: ChainDict) -> List[ChainDict]:
        return [self._id_nodes[_1] for _1 in self.siblings(id(node))]

    def sibling_ancestor(self, node_id: int):
        while self.is_parent_exist(node_id):
            yield self.siblings(node_id) + [node_id, ]
            node_id = self.parent(node_id)
            if not self.is_parent_exist(node_id):
                break # root parent can return itself. 
        if self.exists(node_id) and not self.is_parent_exist(node_id):
            yield [node_id, ]

    def descendants(self, node_id: int) -> List[int]:
        nodes = []
        stack = []
        if self.exists(node_id):
            stack.extend(self.children(node_id))
        while stack:
            pop_node_id = stack.pop()
            nodes.append(pop_node_id)
            stack.extend(self.children(pop_node_id))
        return nodes

    def get_descendants(self, node: ChainDict) -> List[ChainDict]:
        return [self._id_nodes[_1] for _1 in self.descendants(id(node))]

    def ancestor(self, node_id: int):
        while self.is_parent_exist(node_id):
            node_id = self.parent(node_id)
            yield [node_id, ]

    def get_ancestor(self, node: ChainDict) -> List[ChainDict]:
        return [self._id_nodes[_1[0]] for _1 in self.ancestor(id(node))]

    def get_tag(self, node: ChainDict) -> str:
        if id(node) in self._id_tags:
            return self._id_tags[id(node)]
        else:
            return self._id_tags[self._contrast_ids[id(node)]]
    
    def get_attr_names(self, node: ChainDict) -> List[str]:
        attr_names = []
        for reverse_name, node_ids in self._reverse_attr_name.items():
            if id(node) in node_ids:
                attr_names.append(reverse_name[0])
        return attr_names

    def find_text(self, node: ChainDict, format_func: Callable=lambda x:x) -> Any:

        if not isinstance(node, ChainDict):
            raise TextNotFound('Text not found.')
        
        if self.real_cdata_key in node:
            return format_func(node[self.real_cdata_key])
        elif isinstance(node, str):
            return format_func(node)
        else:
            raise TextNotFound('Text not found.')

    def _revert_data(self, forward=True):
        for no, node_ids in self._reverse_attr_name.items():
            new_name, old_name = no
            for node_id in node_ids:
                node = self._id_nodes[node_id]
                if forward:
                    node[new_name] = node[old_name]
                    node.pop(old_name)
                else:
                    node[old_name] = node[new_name]
                    node.pop(new_name)

        for node in self._textids:
            node = self._id_nodes[node]
            if forward:
                node[self.real_cdata_key] = node[self.cdata_key]
                node.pop(self.cdata_key)
            else:
                node[self.cdata_key] = node[self.real_cdata_key]
                node.pop(self.real_cdata_key)



from .manager import ChainManager
from ._ntype import *
import os
from typing import Any, Dict

__all__ = [
    'parse',
    'parse_xml',
    'parse_json',
    'parse_dict',
    'parse_yaml',
    'parse_html_or_xhtml'
]

def parse_xml(
        xml_data: str
        , namespace_separator: str = None
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , analysis_text: bool = True
        , open_cdata: bool = False
        , open_comment: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , analysis_mode: int = AnalysisMode.RECOMMEND
        , combine_cdata: bool = True
        , include_comment: bool = False
        , include_loc: bool = True
    ) -> ChainManager:
    return ChainManager(
        xml_data
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , analysis_text = analysis_text
        , cdata_separator = cdata_separator
        , analysis_mode = analysis_mode
        , combine_cdata = combine_cdata
        , include_comment = include_comment
        , include_loc = include_loc
        , open_cdata = open_cdata
        , open_comment = open_comment
        , namespace_separator = namespace_separator
    )

def parse(
        xml_path: str
        , namespace_separator: str = None
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , encoding: str = 'utf-8'
        , analysis_text: bool = True
        , open_cdata: bool = False
        , open_comment: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , analysis_mode: int = AnalysisMode.RECOMMEND
        , combine_cdata: bool = True
        , include_comment: bool = False
        , include_loc: bool = True
    ) -> ChainManager:
    with open(xml_path, 'r', encoding=encoding) as fp:
        xml_data = fp.read()

    return parse_xml(
        xml_data
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , analysis_text = analysis_text
        , cdata_separator = cdata_separator
        , analysis_mode = analysis_mode
        , combine_cdata = combine_cdata
        , include_comment = include_comment
        , include_loc = include_loc
        , open_cdata = open_cdata
        , open_comment = open_comment
        , namespace_separator = namespace_separator
    )

def parse_json(
        json_data: str
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , analysis_text: bool = True
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
    ) -> ChainManager:

    # if os.path.exists(json_data) and os.path.isfile(json_data):
    #     with open(json_data, 'r', encoding=encoding) as fp:
    #         json_data = fp.read()

    return ChainManager(
        ''
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , data_switch = ParserType.JSON
        , universal_data = json_data
        , analysis_text = analysis_text
        , analysis_mode = analysis_mode
    )

def parse_yaml(
        yaml_data: str
        , encoding='utf-8'
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , analysis_text: bool = True
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
    ) -> ChainManager:

    if os.path.exists(yaml_data) and os.path.isfile(yaml_data):
        with open(yaml_data, 'r', encoding=encoding) as fp:
            yaml_data = fp.read()

    return ChainManager(
        ''
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , data_switch = ParserType.YAML
        , universal_data = yaml_data
        , analysis_text = analysis_text
        , analysis_mode = analysis_mode
    )

def parse_dict(
        dict_data: Dict[str, Any]
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , analysis_text: bool = True
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
    ) -> ChainManager:
    return ChainManager(
        ''
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , data_switch = ParserType.DICT
        , universal_data = dict_data
        , analysis_text = analysis_text
        , analysis_mode = analysis_mode
    )

def parse_html_or_xhtml(html_data: str
        , convert_charrefs: bool = True
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , comment_key: str = COMMENT_KEY
        , include_comment: bool = False
        , open_comment: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , loc_key: str = LOC_KEY
        , include_loc: bool = True
        , analysis_text: bool = True
        , analysis_mode: int = AnalysisMode.RECOMMEND
    ) -> ChainManager:
    
    return ChainManager(
        ''
        , data_switch = ParserType.HTML
        , universal_data = html_data
        , convert_charrefs = convert_charrefs
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , comment_key = comment_key
        , include_comment = include_comment
        , cdata_separator = cdata_separator
        , loc_key = loc_key
        , include_loc = include_loc
        , analysis_text = analysis_text
        , analysis_mode = analysis_mode
        , open_comment = open_comment
    )



HTML = """\
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/3.4.1/css/bootstrap.min.css" integrity="sha384-HSMxcRTRxnN+Bdg0JdbxYKrThecOKuH5zCYotlSAcp1+c8xmyTe9GYg1l9a69psu" crossorigin="anonymous">
    <title>${title}</title>
</head>
<body>
    ${tables}
</body>
</html>
"""

TABLE = """
<table class="table table-bordered">
    <caption></caption>
    <thead>
        ${tr_ths}
    </thead>
    <tbody>
        ${tr_tds}
    </tbody>
    <tfoot></tfoot>
</table>
"""



__all__ = [
    'BaseTable',
]

from abc import ABC, abstractmethod
from typing import Optional
from UniversalParser.html.output import parse_html_table
from UniversalParser._tools import *
from UniversalParser._parse import parse_xml
from UniversalParser._ntype import *
from UniversalParser.micro._ntype import *
from UniversalParser._exception import *
from UniversalParser.manager import ChainManager

class BaseTable(ABC):

    def __init__(self
            , zip_file_path: str
            , tag_type: str
            , read_file_path: Optional[str] = None
            , encoding: str = 'utf-8'
            , mode: str = 'a'
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , real_cdata_key: str = REAL_CDATA_KEY
            , cdata_self_key: str = CDATA_SELF_KEY
            , comment_key: str = COMMENT_KEY
            , analysis_text: bool = False
            , cdata_separator: str = CDATA_SEPARATOR
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
        ) -> None:
        super().__init__()
        self.zip_file_path = zip_file_path
        self.tag_type = tag_type
        self.file_path = read_file_path
        self.encoding = encoding
        self.mode = mode
        self.attr_prefix = attr_prefix
        self._analysis_text = analysis_text
        self.cdata_key = cdata_key
        self.real_cdata_key = real_cdata_key
        self.cdata_self_key = cdata_self_key
        self.comment_key = comment_key
        self.cdata_separator = cdata_separator
        self.analysis_mode = analysis_mode
        self.Tag = self.__get_tag_type(self.tag_type)
        self.tables = [] # all of tables.

        try:
            self.zip_obj = ZipFile(self.zip_file_path, mode=self.mode)
        except Exception as e:
            raise ZipParseError(f'zip-file can not be parsed.{e}')
        self._init_manager()
        self.refresh_tables()

    def _init_manager(self):
        self.ori_xml = read_content_from_zipobj(self.zip_obj, self.file_path, self.encoding)
        self.manager: ChainManager = parse_xml(
            self.ori_xml
            , analysis_text = self._analysis_text
            , attr_prefix = self.attr_prefix
            , cdata_key = self.cdata_key
            , real_cdata_key = self.real_cdata_key
            , cdata_self_key = self.cdata_self_key
            , comment_key = self.comment_key
            , cdata_separator = self.cdata_separator
            , analysis_mode = self.analysis_mode
        )

    @abstractmethod
    def refresh_tables(self): pass

    @abstractmethod
    def modify_cell_by_coordinate(self): pass

    @abstractmethod
    def batch_modify_cells_by_coordinate(self): pass

    def __get_tag_type(self, tag_type: str):
        if MicroTagType.WORD == tag_type:
            return WordTag
        elif MicroTagType.PPT == tag_type:
            return PowerPointTag
        elif MicroTagType.EXCEL == tag_type:
            return ExcelTag
        else:
            raise Exception('Unsupprot doc type.')

    def to_html(self
            , title: str = 'Table'
            , out_path: str = 'tables.html'
            , encoding: str = 'utf-8'
        ):
        with open(out_path, 'w', encoding=encoding) as fp:
            fp.write(parse_html_table(self.tables, title=title))

    def to_word(self): pass

    def to_pdf(self): pass        

    def to_excel(self): pass

    def save_as(self, new_word_path):
        new_xml_data = self.manager.get_xml_data()
        save_docs_with_modify_by_move(self.zip_obj, {self.file_path: new_xml_data}, new_word_path)

    def __del__(self):
        try:
            self.zip_obj.close()
        except:
            pass



from functools import wraps
from ._exception import *

def limit_one(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        result = func(*args, **kwargs)
        if True == kwargs.get('one_'):
            if 1 == len(result):
                return result[0]
            elif 0 == len(result):
                raise NoNodesFound('No match node.')
            else:
                raise MoreNodesFound('More nodes found.')
        else:
            return result
    return decorator



from typing import Any, Dict, Optional, Union
from io import StringIO
from collections import deque
from UniversalParser._exception import *
from UniversalParser._ntype import *

__all__ = [
    'unparse_odict',
]

__author__ = 'jiyang'
__version__ = '0.0.1'
__license__ = 'MIT'

def _init_head(out_stream, xml_declare, encoding: str) -> None:
    if xml_declare is None:
        out_stream.write(f'<?xml version="1.0" encoding="{encoding}"?>\n')
    else:
        version = xml_declare.get('version')
        encoding = xml_declare.get('encoding')
        standalone = xml_declare.get('standalone')

        xml_declare = '<?xml'
        xml_declare += f' version="{version}"'
        if encoding is not None:
            xml_declare += f' encoding="{encoding}"'
        if -1 != standalone:
            real_v = 'yes' if 1==standalone else 'no'
            xml_declare += f' standalone="{real_v}"'
        xml_declare += '?>\n'
        out_stream.write(xml_declare)

def _build_body(out_stream
        , dict_data
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , cdata_separator: str = CDATA_SEPARATOR
        , loc_key: str = LOC_KEY
        , distinct: bool = True
    ) -> None:

    loc_attr_name = f'{attr_prefix}{loc_key}'

    root = list(dict_data.items())[0]
    queue = deque([[root[0], root[1], ''], ])

    while queue:
        pop_obj = queue.popleft()
        pop_tag = pop_obj[0]
        pop_node = pop_obj[1]

        attr_nodes = ''
        comment_nodes = ''
        text_nodes = ''
        cdata_nodes = ''
        sub_nodes = []

        for tag, value in pop_node.items():
            if tag.startswith(attr_prefix):
                if tag != loc_attr_name:
                    attr_nodes += f' {tag[len(attr_prefix):]}="{value}"'
            elif cdata_key == tag:
                text_nodes += f'{value}'
            elif cdata_self_key == tag:
                if distinct: value = set(value)
                for _cdata in value:
                    cdata_nodes += f'<![CDATA[{_cdata}]]>'
            elif comment_key == tag:
                if distinct: value = set(value)
                for _comment in value:
                    comment_nodes += f'<!-- {_comment} -->'
            else:
                if isinstance(value, dict):
                    sub_nodes.append([tag, value, ''])
                elif isinstance(value, (list, tuple)):
                    sub_nodes.extend([[tag, _1, ''] for _1 in value])
                else:
                    raise TypeError('Unknown status')
        
        try:
            sub_nodes.sort(key=lambda x: int(x[1][loc_attr_name]), reverse=False)
        except:
            pass # think of [json, yaml, dict ... ]
        
        if len(sub_nodes) > 0:
            sub_nodes[-1][2] = f'</{pop_tag}>{pop_obj[2]}'
            pop_obj[2] = ''
            out_stream.write(f'<{pop_tag}{attr_nodes}>{comment_nodes}{cdata_nodes}{text_nodes}')
            queue.extendleft(reversed(sub_nodes))
        else:
            if comment_nodes or text_nodes or cdata_nodes:
                out_stream.write(f'<{pop_tag}{attr_nodes}>{comment_nodes}{cdata_nodes}{text_nodes}</{pop_tag}>')
            else:
                out_stream.write(f'<{pop_tag}{attr_nodes}/>')
        if pop_obj[2]:
            out_stream.write(pop_obj[2])

def unparse_dict(
        dict_data: Dict[str, Any]
        , xml_declare: Dict[str, Union[str, int]] = None
        , out_stream = None
        , encoding: str = 'utf-8'
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , cdata_separator: str = CDATA_SEPARATOR
        , loc_key: str = LOC_KEY
        , distinct: bool = True
    ) -> Optional[str]:

    if not (isinstance(dict_data, dict) and 1 == len(dict_data)):
        raise UnparseDictError('Only one root node is allowed.')

    self_stream = False
    if out_stream is None:
        out_stream: StringIO = StringIO()
        self_stream = True

    _init_head(out_stream, xml_declare, encoding)
    _build_body(
        out_stream
        , dict_data
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , cdata_separator = cdata_separator
        , loc_key = loc_key
        , distinct = distinct
    )

    if self_stream:
        return_str = out_stream.getvalue()
        out_stream.close()
        return return_str



from typing import Callable, Optional, Union
import json
from copy import deepcopy
from io import StringIO
from .struct import *
from ._tools import *
from ._decorator import *
from ._collections import *
from ._ntype import *
from UniversalParser.xml import odict as XmlParser
from UniversalParser.xml.unparser import unparse_dict
from UniversalParser.html import odict as HTMLParser

__all__ = [
    'ChainManager',
]

class ChainManager:

    SEARCH_ATTR_KEY: str = 'id'

    def __init__(self
            , xml_data: str # xml
            , *args
            , data_switch: int = ParserType.XML
            , universal_data: str = None # json/dict/yaml/...
            , analysis_text: bool = True
            , open_cdata: bool = False
            , open_comment: bool = False
            , encoding: str = None
            , namespace_separator: str = None
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
            , combine_cdata: bool = True
            , include_comment: bool = False
            , include_loc: bool = False # Compatible with JSON
            , convert_charrefs: bool = True
            , **kwargs
        ) -> None:

        self.data_switch = data_switch # choice analysis obj type, default: ParserType.XML
        self.encoding = encoding
        self.open_comment = open_comment
        self.open_cdata = open_cdata
        self.include_comment = include_comment
        self.combine_cdata = combine_cdata
        self.analysis_text = analysis_text
        self.namespace_separator = namespace_separator
        self.xml_declare = None

        self.attr_prefix = kwargs['attr_prefix'] if 'attr_prefix' in kwargs else ATTR_PREFIX
        self.cdata_key = kwargs['cdata_key'] if 'cdata_key' in kwargs else CDATA_KEY
        self.real_cdata_key = kwargs['real_cdata_key'] if 'real_cdata_key' in kwargs else REAL_CDATA_KEY
        self.cdata_self_key = kwargs['cdata_self_key'] if 'cdata_self_key' in kwargs else CDATA_SELF_KEY
        self.comment_key = kwargs['comment_key'] if 'comment_key' in kwargs else COMMENT_KEY
        self.cdata_separator = kwargs['cdata_separator'] if 'cdata_separator' in kwargs else CDATA_SEPARATOR
        self.loc_key = kwargs['loc_key'] if 'loc_key' in kwargs else LOC_KEY
        ChainDict.cdata_key = self.real_cdata_key
        ChainDict.switch_loc = include_loc
        self.loc_attr = f'{self.attr_prefix}{self.loc_key}'
        
        if ParserType.XML == data_switch:
            temp_data, self.xml_declare = XmlParser.parse_odict(
                xml_data
                , encoding = self.encoding
                , namespace_separator = self.namespace_separator
                , attr_prefix = self.attr_prefix
                , cdata_key = self.cdata_key
                , cdata_self_key = self.cdata_self_key 
                , comment_key = self.comment_key
                , combine_cdata = combine_cdata
                , include_comment = include_comment
                , include_loc = include_loc
                , cdata_separator = self.cdata_separator
            )
        elif ParserType.HTML == data_switch:
            temp_data = HTMLParser.parse_odict(
                universal_data
                , convert_charrefs = convert_charrefs
                , attr_prefix = self.attr_prefix
                , cdata_key = self.cdata_key
                , comment_key = self.comment_key
                , include_comment = self.include_comment
                , cdata_separator = self.cdata_separator
                , loc_key = self.loc_key
                , include_loc = include_loc
            )
        elif ParserType.JSON == data_switch:
            temp_data = json.loads(universal_data)
        elif ParserType.DICT == data_switch:
            temp_data = universal_data
        elif ParserType.YAML == data_switch:
            try:
                import yaml
            except:
                raise ImportError('If you want to use this function, please `pip install yaml`.')
            else:
                temp_data = yaml.load(universal_data, Loader=yaml.FullLoader)
        else:
            raise DocTypeError('Unknown `universal_data` type.')
            
        self.objects = ChainXML(
            temp_data
            , attr_prefix = self.attr_prefix
            , cdata_key = self.cdata_key
            , real_cdata_key = self.real_cdata_key
            , cdata_self_key = self.cdata_self_key
            , comment_key = self.comment_key
            , cdata_separator = self.cdata_separator
            , loc_key = self.loc_key
            , data_switch = data_switch
            , analysis_text = analysis_text
            , analysis_mode = analysis_mode
            , open_cdata = open_cdata
            , open_comment = open_comment
        )

        self.document = self.xml = self.objects.xml

        self._attr_searcher = self.objects._attr_searcher
        self._reverse_attr_name = self.objects._reverse_attr_name
        self._id_nodes = self.objects._id_nodes
        self._id_tags = self.objects._id_tags
        self._tag_ids = self.objects._tag_ids
        self._text_ids = self.objects._text_ids
        self._contrast_ids = self.objects._contrast_ids
        self._comment_ids = self.objects._comment_ids
        self._cdata_ids = self.objects._cdata_ids

        self.get_tag = self.objects.get_tag
        self.get_parent = self.objects.get_parent
        self.get_children = self.objects.get_children
        self.get_siblings = self.objects.get_siblings
        self.get_ancestor = self.objects.get_ancestor
        self.get_descendants = self.objects.get_descendants
        self.get_attr_names = self.objects.get_attr_names
        self.find_text = self.objects.find_text

        self.exists = self.objects.exists
        self.is_parent_exist = self.objects.is_parent_exist

    @limit_one
    def find_nodes_by_attrs(self, one_=False, **kwargs) -> Union[List[ChainDict], ChainDict]:

        if len(kwargs) < 1:
            raise KeywordAttrsError('Keyword parameters must be passed in.')

        conditions = set([f'{k}={v}' for k, v in kwargs.items()])
        match_cons = []
        for sk in list(self._attr_searcher.keys()):
            if sk in conditions:
                match_cons.append(sk)

        if len(match_cons) != len(conditions):
            return []

        len_match_cons = len(match_cons)
        if 0 == len_match_cons:
            return []
        elif 1 == len_match_cons:
            obj = self._attr_searcher[match_cons[0]]
            if obj['count'] > 1:
                return [self._id_nodes[_t] for _t in obj['nodes']]
            else:
                return [self._id_nodes[obj['nodes'][0]], ]
        else:
            objs = find_intersection([self._attr_searcher[t]['nodes'] for t in match_cons])
            if 0 == len(objs):
                return []
            elif 1 == len(objs):
                return [self._id_nodes[objs[0]], ]
            else:
                return [self._id_nodes[_t] for _t in objs]

    def __matmul__(self, attrs: Union[dict, str]) -> ChainExpressionManager: # @
        ChainExpressionManager.SEARCH_ATTR_KEY = self.SEARCH_ATTR_KEY
        if isinstance(attrs, dict):
            nodes = self.find_nodes_by_attrs(**attrs)
        elif isinstance(attrs, str):
            nodes = self.find_nodes_by_attrs(**{self.SEARCH_ATTR_KEY: attrs})
        else:
            raise TypeError('`attrs` must be dict or str.')
        chainExpressionManager = ChainExpressionManager(self, self.objects, nodes)
        return chainExpressionManager

    def find_node_by_attrs(self, **kwargs) -> ChainDict:
        return self.find_nodes_by_attrs(one_=True, **kwargs)

    @limit_one
    def find_nodes_by_indexs(self, indexs: List[int], one_=False, **kwargs) -> Union[List[ChainDict], ChainDict]:
        nodes = self.find_nodes_by_attrs(**kwargs)
        len_nodes = len(nodes)
        return_nodes = []
        for index in indexs:
            if 0 <= index < len_nodes:
                return_nodes.append(nodes[index])
        return return_nodes

    @limit_one
    def find_nodes_by_tag(self, tag: str, one_=False) -> Union[List[ChainDict], ChainDict]:
        nodes = []
        if tag in self._tag_ids:
            for _1 in [self._id_nodes[_id] for _id in self._tag_ids[tag]]:
                if isinstance(_1, ChainDict):
                    nodes.append(_1)
                elif isinstance(_1, list):
                    pass
        return nodes

    def __or__(self, tag) -> ChainExpressionManager: # |
        nodes = self.find_nodes_by_tag(tag)
        chainExpressionManager = ChainExpressionManager(self, self.objects, nodes)
        return chainExpressionManager

    @limit_one
    def find_nodes_by_comment(self, comment: str, one_=False) -> Union[List[ChainDict], ChainDict]:
        nodes = []
        if comment in self._comment_ids:
            for _1 in [self._id_nodes[_id] for _id in set(self._comment_ids[comment])]:
                nodes.append(_1)
        return nodes

    def __floordiv__(self, comment) -> ChainExpressionManager: # //
        nodes = self.find_nodes_by_comment(comment)
        chainExpressionManager = ChainExpressionManager(self, self.objects, nodes)
        return chainExpressionManager

    @limit_one
    def find_nodes_by_cdata(self, cdata: str, one_=False) -> Union[List[ChainDict], ChainDict]:
        nodes = []
        if cdata in self._cdata_ids:
            for _1 in [self._id_nodes[_id] for _id in set(self._cdata_ids[cdata])]:
                nodes.append(_1)
        return nodes

    def __mod__(self, cdata) -> ChainExpressionManager: # %
        nodes = self.find_nodes_by_cdata(cdata)
        chainExpressionManager = ChainExpressionManager(self, self.objects, nodes)
        return chainExpressionManager
    
    @limit_one
    def find_nodes_by_tag_and_attrs(self, tag_: str=None, one_=False, **kwargs) -> Union[List[ChainDict], ChainDict]:

        if tag_ is not None and tag_ not in self._tag_ids:
            return []

        conditions = set([f'{k}={v}' for k, v in kwargs.items()])
        match_cons = []
        for sk in list(self._attr_searcher.keys()):
            if sk in conditions:
                match_cons.append(sk)

        if len(match_cons) != len(conditions):
            if 0 != len(conditions):
                return []
                
        len_match_cons = len(match_cons)
        if 0 == len_match_cons and 0 != len(conditions):
            return []
        else:
            if tag_ is not None:
                tag_ids = [id(_1) for _1 in self.find_nodes_by_tag(tag_)]
                objs = find_intersection([tag_ids] + [self._attr_searcher[t]['nodes'] for t in match_cons])
            else:
                objs = find_intersection([self._attr_searcher[t]['nodes'] for t in match_cons])

            if 0 == len(objs):
                return []
            elif 1 == len(objs):
                return [self._id_nodes[objs[0]], ]
            else:
                return [self._id_nodes[_t] for _t in objs]

    @limit_one
    def find_nodes_by_tag_text_attrs(self, tag_: str=None, text_: str=None, one_=False, **kwargs) -> Union[List[ChainDict], ChainDict]:
        tag_ids = None
        text_ids = None
        attr_ids = None

        if tag_ is not None:
            tag_ids = [id(_1) for _1 in self.find_nodes_by_tag(tag_)]
        if text_ is not None:
            text_ids = [id(_1) for _1 in self.find_nodes_by_text(text_)]
        if len(kwargs) > 0:
            attr_ids = [id(_1) for _1 in self.find_nodes_by_attrs(**kwargs)]

        node_ids = find_intersection([_1 for _1 in (tag_ids, text_ids, attr_ids) if _1 is not None])
        return [self._id_nodes[node_id] for node_id in node_ids]

    @limit_one
    def find_nodes_by_text(self, text: str, one_=True) -> Union[List[ChainDict], ChainDict]:
        if not self.analysis_text:
            raise EnvironmentError('If you want to use this way, please set `analysis_text=True`')
        if text in self._text_ids:
            return [self._id_nodes[_1] for _1 in self._text_ids[text]]
        return []

    def __truediv__(self, text) -> ChainExpressionManager: # /
        nodes = self.find_nodes_by_text(text)
        chainExpressionManager = ChainExpressionManager(self, self.objects, nodes)
        return chainExpressionManager

    def find_text_by_attrs(self, *args, text_type: int=TextType.STR, format_func: Callable=lambda x:x, **kwargs) -> str:
        node = self.find_nodes_by_attrs(*args, one_=True, **kwargs)
        text = self.find_text(node, format_func)
        if TextType.STR == text_type:
            return str(text)
        elif TextType.FLOAT == text_type:
            return float(text)
        elif TextType.INT == text_type:
            return int(text)
        else:
            return text

    def __find_constraint_node(self
            , unique_node: Optional[ChainDict] = None
            , find_func: Optional[Callable] = None
            , constraint: Optional[Dict[str, Any]] = None
        ) -> int:
        
        unique_node_id = None
        if unique_node is not None:
            if isinstance(unique_node, ChainDict):
                unique_node_id = id(unique_node)
            else:
                raise KeywordAttrsError('`unique_node` must be type of ChainDict.')

        if unique_node_id is None:
            if constraint is not None and isinstance(constraint, dict):
                if 'args' not in constraint or 'kwargs' not in constraint:
                    raise KeywordAttrsError('`args` and `kwargs` must in constraint.')
                find_func = find_func if find_func is not None else self.find_nodes_by_tag_text_attrs
                node = find_func(*constraint['args'], **constraint['kwargs'], one_=True)
                unique_node_id = id(node)
            else:
                raise KeywordAttrsError('`constraint` must be assign.')

        if unique_node_id is None:
            raise NoNodesFound('Constraint node not found.')

        return unique_node_id

    @limit_one
    def find_nodes_with_sibling_ancestor(self
            , unique_node: Optional[ChainDict] = None
            , *args
            , tag_: str = None
            , text_: str = None
            , one_=False
            , find_func: Optional[Callable] = None # default: find_nodes_by_tag_text_attrs
            , constraint: Optional[Dict[str, Any]] = None # {'args': [], 'kwargs': {}}
            , **kwargs
        ) -> Union[List[ChainDict], ChainDict]:

        unique_node_id = self.__find_constraint_node(unique_node, find_func, constraint)

        t_node_ids = [id(_1) for _1 in self.find_nodes_by_tag_text_attrs(tag_=tag_, text_=text_, *args, **kwargs)]
        if len(t_node_ids) <= 0:
            return []

        nodes = []
        for t_node_id in t_node_ids:
            for _1 in self.objects.sibling_ancestor(t_node_id):
                if unique_node_id in _1:
                    nodes.append(self._id_nodes[t_node_id])
                    break
        return nodes

    @limit_one
    def find_nodes_with_ancestor(self
            , unique_node: Optional[ChainDict] = None
            , *args
            , tag_: str = None
            , text_: str = None
            , one_=False
            , find_func: Optional[Callable] = None # default: find_nodes_by_tag_text_attrs
            , constraint: Optional[Dict[str, Any]] = None # {'args': [], 'kwargs': {}}
            , **kwargs
        ) -> Union[List[ChainDict], ChainDict]:

        unique_node_id = self.__find_constraint_node(unique_node, find_func, constraint)

        t_node_ids = [id(_1) for _1 in self.find_nodes_by_tag_text_attrs(tag_=tag_, text_=text_, *args, **kwargs)]
        if len(t_node_ids) <= 0:
            return []

        nodes = []
        for t_node_id in t_node_ids:
            for _1 in self.objects.ancestor(t_node_id):
                if unique_node_id in _1:
                    nodes.append(self._id_nodes[t_node_id])
                    break
        return nodes

    @limit_one
    def find_nodes_with_descendants(self
            , unique_node: Optional[ChainDict] = None
            , *args
            , tag_: str = None
            , text_: str = None
            , one_=False
            , find_func: Optional[Callable] = None # default: find_nodes_by_tag_text_attrs
            , constraint: Optional[Dict[str, Any]] = None # {'args': [], 'kwargs': {}}
            , **kwargs
        ) -> Union[List[ChainDict], ChainDict]:
        
        unique_node_id = self.__find_constraint_node(unique_node, find_func, constraint)

        t_node_ids = [id(_1) for _1 in self.find_nodes_by_tag_text_attrs(tag_=tag_, text_=text_, *args, **kwargs)]
        if len(t_node_ids) <= 0:
            return []

        nodes = []
        for t_node_id in t_node_ids:
            if unique_node_id in self.objects.descendants(t_node_id):
                nodes.append(self._id_nodes[t_node_id])
        return nodes

    def popitem(self, obj: ChainDict):
        self.objects.signal_del_node_base(obj)

    def pop_node_by_attrs(self, **kwargs):
        obj = self.find_nodes_by_attrs(one_=True, **kwargs)
        return self.popitem(obj)

    def pop_nodes_by_attrs(self, **kwargs):
        objs = self.find_nodes_by_attrs(**kwargs)
        for obj in objs:
            self.popitem(obj)

    def insert_attrs(self, node, *args, **kwargs) -> None:
        for attr_name, value in args+tuple(kwargs.items()):
            if attr_name not in node:
                node[attr_name] = value
                self.objects.signal_add_attr_base(node, attr_name, value)

    def insert_comment(self, node: ChainDict, comment: str) -> None:
        if not isinstance(node, ChainDict):
            raise TypeError('Only support `ChainDict` node.')
        if self.open_comment and self.include_comment:
            if self.comment_key in node:
                node[self.comment_key].append(comment)
            else:
                node[self.comment_key] = [comment, ]
            self.objects.registry_comment_ids(comment, node)
        else:
            raise EnvironmentError('Must set `open_comment=True` and `include_comment=True`.')

    def insert_cdata(self, node: ChainDict, cdata: str):
        if not isinstance(node, ChainDict):
            raise TypeError('Only support `ChainDict` node.')
        if not self.combine_cdata and self.open_cdata:
            if self.cdata_self_key in node:
                node[self.cdata_self_key].append(cdata)
            else:
                node[self.cdata_self_key] = [cdata, ]
            self.objects.registry_cdata_ids(cdata, node)
        else:
            raise EnvironmentError('Must set `combine_cdata=False` and `open_cdata=True`.')

    def update_attr(self, node: ChainDict, attr_name: str, new_value: Any) -> Tuple[str, str]:
        if attr_name in node:
            old_search_key = f'{attr_name}={node[attr_name]}'
            return_v = (attr_name, node[attr_name])
            node[attr_name] = new_value
            new_search_key = f'{attr_name}={new_value}'
            self.objects.signal_modify_attr(old_search_key, new_search_key, node)
            return return_v
        else:
            raise UpdateError('Unable to update this node, check whether the attribute exists.')

    def del_attr(self, node: ChainDict, attr_name: str) -> Tuple[str, str]:
        attr_value = node[attr_name]
        return_v = (attr_name, attr_value)
        node.pop(attr_name)
        self.objects.signal_del_attr_base(node, attr_name, attr_value)
        return return_v

    def batch_del_attr(self, node: ChainDict, *args) -> List[Tuple[str, str]]:
        return_vs = []
        for attr_name in args:
            return_vs.append(self.del_attr(node, attr_name))
        return return_vs

    def batch_update_attrs(self, node, *args, **kwargs) -> List[Tuple[str, str]]:
        if len(args) < 1 and len(kwargs) < 1:
            raise KeywordAttrsError('Parameters or keyword parameters must be passed in.')
        return_vs = []
        for attr_name, value in args+tuple(kwargs.items()):
            return_vs.append(self.update_attr(node, attr_name, value))
        return return_vs
        
    def update_text(self, node: ChainDict, new_text: Any) -> str:
        if self.real_cdata_key in node:
            old_text = node[self.real_cdata_key]
            node[self.real_cdata_key] = new_text
            self.objects.signal_modify_text(old_text, new_text, node)
            return old_text
        else:
            raise UpdateError('Unable to update this node, `text_` is not found.')

    def clear_text(self, node: ChainDict) -> str:
        return self.update_text(node, '')

    def batch_clear_text(self, nodes: List[ChainDict]) -> List[str]:
        old_values = []
        for node in nodes:
            old_values.append(self.update_text(node, ''))
        return old_values

    def _reset_locs(self, node: Union[ChainDict, List[ChainDict]]) -> None:
        start = 0
        nodes = []
        if isinstance(node, ChainDict):
            nodes = self.get_children(node)
        elif isinstance(node, list):
            parent_node = self.get_parent(node)
            nodes = self.get_children(parent_node)
        nodes.sort(key=lambda x: int(x[self.loc_key]), reverse=False)
        for t_node in nodes:
            self.update_attr(t_node, self.loc_key, str(start))
            start += 1

    def _insert_list(self, node: list, tag: str='', attrs={}, text: str=''):
        inner_node = ChainDict(self.objects)
        self.insert_attrs(inner_node, **attrs)
        self.insert_attrs(inner_node, **{self.loc_key: self.objects._get_last_loc(node)})
        inner_node[self.real_cdata_key] = text
        node.append(inner_node)
        self.objects.signal_add_node(inner_node, node, tag=tag, text='')
        return inner_node
                
    def insert(self, node: Union[ChainDict, List[ChainDict]], tag: str='', attrs={}, text: str='') -> ChainDict:
        if isinstance(node, ChainDict):
            if tag not in node:
                insert_node = ChainDict(self.objects)
                insert_node[self.real_cdata_key] = ''
                inner_node = ChainDict(self.objects)
                self.insert_attrs(inner_node, **attrs)
                self.insert_attrs(inner_node, **{self.loc_key: self.objects._get_last_loc(node)})
                inner_node[self.real_cdata_key] = text
                insert_node[tag] = inner_node
                node.update(insert_node)
                self.objects.signal_add_node(inner_node, insert_node, tag=tag, text=text)
                self.objects.signal_add_node(insert_node, node, tag=tag, text='')
                return inner_node
            else:
                temp_node = node[tag]
                if isinstance(temp_node, ChainDict):
                    temp_list = []
                    node[tag] = temp_list
                    temp_list.append(temp_node)
                    self.objects.signal_add_node(temp_list, node, tag=tag)
                    self.objects.signal_move_node(temp_node, from_node=node, to_node=temp_list)
                    return self._insert_list(temp_list, tag=tag, attrs=attrs, text=text)
                elif isinstance(temp_node, list):
                    return self._insert_list(temp_node, tag=tag, attrs=attrs, text=text)
        elif isinstance(node, list):
            if self.get_tag(node) != tag:
                raise AddNodeError('`tag` error.')
            return self._insert_list(node, tag=tag, attrs=attrs, text=text)
        else:
            raise AddNodeError('Add node error, obj must be `ChainDict` type or `List` type.')
        self._reset_locs(node)

    def move(self, move_node, from_node, to_node) -> None:
        tag = self.get_tag(move_node)
        if id(from_node) == id(to_node):
            raise MoveError("`from_node` can't be the same as `to_node`.")
        if tag not in from_node:
            raise NoNodesFound('`from_node` error.')
        else:
            if isinstance(from_node[tag], ChainDict):
                from_node.pop(tag)
            else:
                from_node[tag].remove(move_node)
                from_node = from_node[tag]
            
            if isinstance(to_node, ChainDict):
                if tag not in to_node:
                    self.update_attr(move_node, self.loc_key, self.objects._get_last_loc(to_node))
                    to_node[tag] = move_node
                else:
                    if isinstance(to_node[tag], ChainDict):
                        temp_node = to_node[tag]
                        temp_list = []
                        to_node[tag] = temp_list
                        temp_list.append(temp_node)
                        self.objects.signal_add_node(temp_list, to_node, tag=tag)
                        self.objects.signal_move_node(temp_node, from_node=to_node, to_node=temp_list)
                        temp_list.append(move_node)
                        to_node = temp_list
                    else:
                        self.update_attr(move_node, self.loc_key, self.objects._get_last_loc(to_node[tag]))
                        to_node[tag].append(move_node)
            else:
                self.update_attr(move_node, self.loc_key, self.objects._get_last_loc(to_node))
                to_node.append(move_node)
                
        self.objects.signal_move_node(move_node, from_node=from_node, to_node=to_node)
        self._reset_locs(from_node)
        self._reset_locs(to_node)

    def copy_to(self, aim_node: Union[ChainDict], node_data: Union[ChainDict, List[ChainDict]]) -> None:
        if isinstance(node_data, list):
            for _data in node_data:
                self.objects._revert_data(False)
                temp_node = deepcopy(_data)
                self.objects._revert_data()
                datas = {self.get_tag(_data): temp_node}
                self.objects._build_by_copy(datas, aim_node)
        else:
            self.objects._revert_data(False)
            temp_node = deepcopy(node_data)
            self.objects._revert_data()
            datas = {self.get_tag(node_data): temp_node}
            self.objects._build_by_copy(datas, aim_node)
        
        self._reset_locs(aim_node)

    def insert_xmlstring(self
            , aim_node: ChainDict
            , xml_str: str
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , real_cdata_key: str = REAL_CDATA_KEY
            , cdata_self_key: str = CDATA_SELF_KEY
            , comment_key: str = COMMENT_KEY
            , analysis_text: bool = True
            , open_cdata: bool = False
            , open_comment: bool = False
            , cdata_separator: str = CDATA_SEPARATOR
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
            , combine_cdata: bool = True
            , include_comment: bool = False
            , include_loc: bool = True
        ) -> None:
        if not isinstance(aim_node, ChainDict):
            raise TypeError('Only support `ChainDict` type.')
        manager = _parse_xml(
            xml_str
            , attr_prefix = attr_prefix
            , cdata_key = cdata_key
            , real_cdata_key = real_cdata_key
            , cdata_self_key = cdata_self_key
            , comment_key = comment_key
            , analysis_text = analysis_text
            , open_cdata = open_cdata
            , open_comment = open_comment
            , cdata_separator = cdata_separator
            , analysis_mode = analysis_mode
            , combine_cdata = combine_cdata
            , include_comment = include_comment
            , include_loc = include_loc
        )
        manager.objects._revert_data(False)
        temp_node = deepcopy(manager.xml)
        manager.objects._revert_data()
        root = list(temp_node.items())[0] # Safety, onle one root node.
        datas = {root[0]: root[1]}
        self.objects._build_by_copy(datas, aim_node)
        self._reset_locs(aim_node)

    def clear_node_content(self, node: ChainDict) -> None:
        if not isinstance(node, ChainDict):
            raise TypeError('Only support `ChainDict` type.')
        self.clear_text(node)
        for ch in self.get_children(node):
            self.popitem(ch)
    
    def clear_node_attrs(self, node: ChainDict) -> None:
        if not isinstance(node, ChainDict):
            raise TypeError('Only support `ChainDict` type.')
        for attr_name in self.get_attr_names(node):
            self.del_attr(node, attr_name)

    def clear_node(self, node: ChainDict) -> None:
        self.clear_node_content(node)
        self.clear_node_attrs(node)

    def get_xml_data(self) -> str:
        self.objects._revert_data(forward=False)
        output = StringIO()
        unparse_dict(
            self.xml
            , xml_declare = self.xml_declare
            , out_stream = output
            , encoding = self.encoding
            , attr_prefix = self.attr_prefix
            , cdata_key = self.cdata_key
            , cdata_self_key = self.cdata_self_key
            , comment_key = self.comment_key
            , cdata_separator = self.cdata_separator
            , loc_key = self.loc_key
        )
        self.objects._revert_data()
        return output.getvalue()

    def save_as_xml(self, path: str = 'output.xml', encoding='utf-8', distinct: bool=True) -> None:
        self.objects._revert_data(forward=False)
        with open(path, 'w', encoding=encoding) as fp:
            unparse_dict(
                self.xml
                , xml_declare = self.xml_declare
                , out_stream = fp
                , encoding = self.encoding
                , attr_prefix = self.attr_prefix
                , cdata_key = self.cdata_key
                , cdata_self_key = self.cdata_self_key
                , comment_key = self.comment_key
                , cdata_separator = self.cdata_separator
                , loc_key = self.loc_key
                , distinct = distinct
            )
        self.objects._revert_data()

    def save_as_json(self
            , path: str = 'output.json'
            , encoding = 'utf-8'
            , ensure_ascii: bool = True
            , indent = None
            , ori = False
        ) -> None:
        if ori:
            self.objects._revert_data(forward=False)
        with open(path, 'w', encoding=encoding) as fp:
            json.dump(self.xml, fp, ensure_ascii=ensure_ascii, indent=indent)
        if ori:
            self.objects._revert_data()

def _parse_xml(
        xml_data: str
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , analysis_text: bool = True
        , open_cdata: bool = True
        , open_comment: bool = True
        , cdata_separator: str = CDATA_SEPARATOR
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
        , combine_cdata: bool = True
        , include_comment: bool = False
        , include_loc: bool = True
    ) -> ChainManager:
    return ChainManager(
        xml_data
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , analysis_text = analysis_text
        , cdata_separator = cdata_separator
        , analysis_mode = analysis_mode
        , combine_cdata = combine_cdata
        , include_comment = include_comment
        , include_loc = include_loc
        , open_cdata = open_cdata
        , open_comment = open_comment
    )



from html.parser import HTMLParser
from collections import OrderedDict
from typing import List
from UniversalParser._ntype import *
from ._ntype import *

class UPHTMLParser(HTMLParser):

    def __init__(self, *
            , convert_charrefs: bool
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , comment_key: str = COMMENT_KEY
            , include_comment: bool = False
            , cdata_separator: str = CDATA_SEPARATOR
            , loc_key: str = LOC_KEY
            , include_loc: bool = False
        ) -> None:
        super().__init__(convert_charrefs=convert_charrefs)

        self.attr_prefix = attr_prefix
        self.cdata_key = cdata_key
        self.cdata_separator = cdata_separator
        self.include_comment = include_comment
        self.comment_key = comment_key
        self.include_loc = include_loc
        self.loc_key = loc_key

        self.documentation: OrderedDict = OrderedDict({})
        self.stack_nodes: List[OrderedDict] = [[self.documentation, [], 0], ]
        self.temp_loc: int = 0
        self.flag_out: int = 0

    def handle_starttag(self, tag, attrs):
        if 0 == self.flag_out: # only in.
            self.temp_loc = 0
        current_node = self.stack_nodes[-1][0]
        node = OrderedDict({})
        self.stack_nodes.append([node, [], self.temp_loc])
        for attr_name, attr_value in attrs:
            node[f'{self.attr_prefix}{attr_name}'] = attr_value
        if self.include_loc:
            node[f'{self.attr_prefix}{self.loc_key}'] = f'{self.temp_loc}'
        if tag in current_node:
            sub_node = current_node[tag]
            if isinstance(sub_node, OrderedDict):
                current_node[tag] = [sub_node, node, ]
            elif isinstance(sub_node, list):
                current_node[tag].append(node)
        else:
            current_node.update(OrderedDict({tag: node}))
        self.flag_out = 0

        if tag in HTML_SINGLE_TAGS:
            self.handle_endtag(tag)

    def handle_endtag(self, tag):
        node = self.stack_nodes.pop()
        node[0][self.cdata_key] = self.cdata_separator.join(node[1])

        self.flag_out = 1
        self.temp_loc = f'{int(node[2])+1}'

    def handle_startendtag(self, tag, attrs):
        if 0 == self.flag_out:
            self.temp_loc = 0
        current_node = self.stack_nodes[-1][0]
        node = OrderedDict({})
        self.stack_nodes.append([node, [], self.temp_loc])
        for attr_name, attr_value in attrs:
            node[f'{self.attr_prefix}{attr_name}'] = attr_value
        if self.include_loc:
            node[f'{self.attr_prefix}{self.loc_key}'] = f'{self.temp_loc}'
        if tag in current_node:
            sub_node = current_node[tag]
            if isinstance(sub_node, OrderedDict):
                current_node[tag] = [sub_node, node, ]
            elif isinstance(sub_node, list):
                current_node[tag].append(node)
        else:
            current_node.update(OrderedDict({tag: node}))
        self.flag_out = 0
        self.handle_endtag(tag)

    def handle_data(self, data):
        t_data = data.strip()
        if t_data:
            self.stack_nodes[-1][1].append(t_data)

    # def handle_entityref(self, name):
    #     ...

    def handle_comment(self, data):
        if self.include_comment:
            node = self.stack_nodes[-1][0]
            if self.comment_key in node:
                node[self.comment_key].append(data.strip())
            else:
                node[self.comment_key] = [data.strip(), ]

def parse_odict(html_data: str
        , convert_charrefs: bool = True
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , comment_key: str = COMMENT_KEY
        , include_comment: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , loc_key: str = LOC_KEY
        , include_loc: bool = False
    ) -> OrderedDict:

    parser = UPHTMLParser(convert_charrefs=convert_charrefs
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , comment_key = comment_key
        , include_comment = include_comment
        , cdata_separator = cdata_separator
        , loc_key = loc_key
        , include_loc = include_loc
    )
    parser.feed(html_data)
    html = parser.documentation
    parser.close()
    return html



from ._parse import *
from ._ntype import *
from ._exception import *
from ._rely import *
from ._collections import *
from .manager import *
from .struct import *

# shotcut
from . import xml as XmlParser
from .html import odict as HtmlParser
from .micro import Excel as ExcelParser
from .micro import Word as WordParser
from .micro import PowerPoint as PowerPointParser

name = 'UniversalParser'
__version__ = '0.2.7.0'
__author__ = 'jiyang'
__license__ = 'MIT'



__all__ = [
    'parse_tables',
]

from typing import Any, Dict, Tuple
from UniversalParser._ntype import *
from UniversalParser.micro._ntype import *
from .BaseTable import *
from UniversalParser._tools import *

class WordTable(BaseTable):

    def __init__(self
            , zip_file_path: str
            , read_file_path: str = 'word/document.xml'
            , encoding: str = 'utf-8'
            , mode: str = 'a'
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , real_cdata_key: str = REAL_CDATA_KEY
            , cdata_self_key: str = CDATA_SELF_KEY
            , comment_key: str = COMMENT_KEY
            , analysis_text: bool = False
            , cdata_separator: str = CDATA_SEPARATOR
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
        ) -> None:

        super().__init__(zip_file_path
            , MicroTagType.WORD
            , read_file_path = read_file_path
            , encoding = encoding
            , mode = mode
            , analysis_text = analysis_text
            , attr_prefix = attr_prefix
            , cdata_key = cdata_key
            , real_cdata_key = real_cdata_key
            , cdata_self_key = cdata_self_key
            , comment_key = comment_key
            , cdata_separator = cdata_separator
            , analysis_mode = analysis_mode
        )

    def refresh_tables(self):
        self.tables = []
        for w_tbl in self.manager.find_nodes_by_tag(
                self.Tag.TABLE
            ):
            table = []
            for w_tr in self.manager.find_nodes_with_ancestor(
                    w_tbl
                    , tag_ = self.Tag.TR
                ):
                row = []
                for w_tc in self.manager.find_nodes_with_ancestor(
                        w_tr
                        , tag_ = self.Tag.TC
                    ):
                    try:
                        row.append(
                            self.manager.find_text(
                                w_tc[self.Tag.P][self.Tag.R][self.Tag.TEXT]
                            )
                        )
                    except:
                        row.append('')
                table.append(row)
            self.tables.append(table)
        return self.tables

    def modify_cell_by_coordinate(self, table_loc: int, coordinate: Tuple[ROW, COL], new_value: Any) -> Any:
        row, col = coordinate
        try:
            table = self.manager.find_nodes_by_tag(self.Tag.TABLE)[table_loc-1]
            row_obj = self.manager.find_nodes_with_ancestor(table, tag_ = self.Tag.TR)[row-1]
            cell_obj = self.manager.find_nodes_with_ancestor(row_obj, tag_ = self.Tag.TC)[col-1]
            tag_P = cell_obj[self.Tag.P]
        except Exception as e:
            raise ZipParseError(f'Modify zip-file error.{e}')
        else:
            try: # MicroSoft Word 2010 will take apart illegal word. Combine them.
                old_value = self.manager.update_text(tag_P[self.Tag.R][self.Tag.TEXT], new_value)
            except: # Unfound
                old_value = ''
                tag_R = self.manager.insert(tag_P, self.Tag.R, {})
                tag_RPR = self.manager.insert(tag_R, self.Tag.RPR, {})
                self.manager.insert(tag_RPR, self.Tag.RFONTS, {self.Tag.Attribute.HINT: 'eastAsia'})
                self.manager.insert(tag_RPR, self.Tag.VERTALIGN, {self.Tag.Attribute.VAL: 'baseline'})
                self.manager.insert(tag_RPR, self.Tag.LANG, {self.Tag.Attribute.VAL: 'en-US', self.Tag.Attribute.EASTASIA: 'zh-CN'})
                self.manager.insert(tag_R, self.Tag.TEXT, {}, new_value)
            return old_value

    def batch_modify_cells_by_coordinate(self
            , table_loc: int
            , coordinate_values: Dict[Tuple[ROW, COL], Any]
        ) -> Dict[Tuple[ROW, COL], Any]:
        old_values = {}
        for coordinate, new_value in coordinate_values.items():
            old_values[coordinate] = self.modify_cell_by_coordinate(table_loc, coordinate, new_value)
        return old_values
    
    def save_from_modify(self, new_word_path: str='output.docx', patt=None, **kwargs):
        new_content = patt_template_replace(self.ori_xml, patt=patt, **kwargs)
        save_docs_with_modify_by_move(self.zip_obj, {self.file_path: new_content}, new_word_path)

    
def parse_tables(
        zip_file_path: str
        , read_file_path: str = 'word/document.xml'
        , encoding: str = 'utf-8'
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , analysis_text: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
    ) -> WordTable:
    return WordTable(
        zip_file_path = zip_file_path
        , read_file_path = read_file_path
        , encoding = encoding
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , analysis_text = analysis_text
        , cdata_separator = cdata_separator
        , analysis_mode = analysis_mode
    )



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="UniversalParser",
    version="0.2.7.0",
    author="jiyang",
    author_email="jiyangj@foxmail.com",
    description="一款通用的文本格式解析器，支持XML、JSON、YAML、HTML、XHTML、Word、Excel、PPT等文本格式的快速解析，在方法上具有高度一致性。",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitee.com/jiyangj/universal-parser",
    packages=setuptools.find_packages(),
    install_requires=[],
    python_requires='>=3.5',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


class DocTypeError(Exception): pass

class KeywordAttrsError(Exception): pass

class MoreNodesFound(Exception): pass

class NoNodesFound(Exception): pass

class TextNotFound(Exception): pass

class UpdateError(Exception): pass

class AddNodeError(Exception): pass

class PopError(Exception): pass

class ZipParseError(Exception): pass

class AnalysisTypeError(Exception): pass

class UnparseDictError(Exception): pass

class ExcelCellLocError(Exception): pass

class MoveError(Exception): pass

class EnvironmentError(Exception): pass



__all__ = [
    'parse_tables',
]

from typing import Any
import re
from .BaseTable import *
from UniversalParser._tools import *
from UniversalParser._parse import parse_xml
from UniversalParser._ntype import *
from UniversalParser.micro._ntype import *
from UniversalParser.manager import ChainManager

class ExcelTable(BaseTable):
    
    def __init__(self
            , zip_file_path: str
            , encoding: str = 'utf-8'
            , mode: str = 'a'
            , attr_prefix: str = ATTR_PREFIX
            , cdata_key: str = CDATA_KEY
            , real_cdata_key: str = REAL_CDATA_KEY
            , cdata_self_key: str = CDATA_SELF_KEY
            , comment_key: str = COMMENT_KEY
            , analysis_text: bool = False
            , cdata_separator: str = CDATA_SEPARATOR
            , analysis_mode: int = AnalysisMode.RECURSION_OLD
        ) -> None:

        super().__init__(zip_file_path
            , MicroTagType.EXCEL
            , encoding = encoding
            , mode = mode
            , analysis_text = analysis_text
            , attr_prefix = attr_prefix
            , cdata_key = cdata_key
            , real_cdata_key = real_cdata_key
            , cdata_self_key = cdata_self_key
            , comment_key = comment_key
            , cdata_separator = cdata_separator
            , analysis_mode = analysis_mode
        )

    def _init_manager(self):
        self.ori_xmls = []
        self.managers: List[ChainManager] = []
        self.file_paths = []
        for info in self.zip_obj.infolist():
            sp_filename = info.filename.split('/')
            if not info.is_dir() and 3==len(sp_filename) and ['xl', 'worksheets']==sp_filename[:2] and sp_filename[2].endswith('.xml'):
                self.file_paths.append(info.filename)
        self.file_paths.sort(key=lambda x: x.replace('sheet','').replace('.xml',''), reverse=False)
        for filename in self.file_paths:
            ori_xml = read_content_from_zipobj(self.zip_obj, filename, self.encoding)
            self.ori_xmls.append(ori_xml)
            self.managers.append(parse_xml(
                ori_xml
                , analysis_text = self._analysis_text
                , attr_prefix = self.attr_prefix
                , cdata_key = self.cdata_key
                , real_cdata_key = self.real_cdata_key
                , cdata_self_key = self.cdata_self_key
                , comment_key = self.comment_key
                , cdata_separator = self.cdata_separator
                , analysis_mode = self.analysis_mode
            ))

    def refresh_tables(self): pass

    def modify_cell_by_coordinate(self, sheet_number: int, coordinate: str, new_value: Any) -> str:
        manager: ChainManager = self.managers[sheet_number-1]
        try:
            node = manager.find_node_by_attrs(**{self.Tag.Attribute.LOC: coordinate})
        except:
            patt_r = re.match(r'^([A-Z]+)([0-9]+)$', coordinate)
            if patt_r:
                col = patt_r.group(1)
                row = str(patt_r.group(2))
                try:
                    row_node = manager.find_nodes_by_tag_and_attrs(tag_=self.Tag.TR, **{self.Tag.Attribute.LOC:row})
                except:
                    table_node = manager.find_nodes_by_tag(self.Tag.TABLE, one_=True)
                    row_node = manager.insert(table_node, 'row', **{})
                
            else:
                raise ExcelCellLocError('Unable to locate excel cell.')
        return manager.update_text(node[self.Tag.TEXT], new_value)

    def batch_modify_cells_by_coordinate(self, sheet_number: int, coordinate_values: Dict[str, Any]) -> List[str]:
        old_values = []
        for coordinate, new_value in coordinate_values.items():
            old_values.append(self.modify_cell_by_coordinate(sheet_number, coordinate, new_value))
        return old_values

    def save_as(self, new_word_path):
        struct_datas = {self.file_paths[_1]:self.managers[_1].get_xml_data() for _1 in range(len(self.file_paths))}
        save_docs_with_modify_by_move(self.zip_obj, struct_datas, new_word_path)

def parse_tables(
        zip_file_path: str
        , encoding: str = 'utf-8'
        , attr_prefix: str = ATTR_PREFIX
        , cdata_key: str = CDATA_KEY
        , real_cdata_key: str = REAL_CDATA_KEY
        , cdata_self_key: str = CDATA_SELF_KEY
        , comment_key: str = COMMENT_KEY
        , analysis_text: bool = False
        , cdata_separator: str = CDATA_SEPARATOR
        , analysis_mode: int = AnalysisMode.RECURSION_OLD
    ) -> ExcelTable:

    return ExcelTable(
        zip_file_path = zip_file_path
        , encoding = encoding
        , attr_prefix = attr_prefix
        , cdata_key = cdata_key
        , real_cdata_key = real_cdata_key
        , cdata_self_key = cdata_self_key
        , comment_key = comment_key
        , analysis_text = analysis_text
        , cdata_separator = cdata_separator
        , analysis_mode = analysis_mode
    )



from collections import OrderedDict
from typing import Any, Union
from ._tools import find_intersection
from ._ntype import SM
from functools import reduce

class ChainDict(OrderedDict):

    cdata_key = 'text_'
    switch_loc = True

    def __init__(self, objects=None):
        super(ChainDict, self).__init__()
        self.__objects__ = objects

    def __node_num(self):
        if self.switch_loc:
            return 2
        else:
            return 1

    def __getattr__(self, name):
        if not name.startswith('_'):
            if isinstance(self[name], dict) and self.__node_num() == len(self[name]) and self.cdata_key in self[name]:
                return self[name][self.cdata_key]
            else:
                return self[name]
        return super(ChainDict, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name == self.cdata_key:
            pass
        if not name.startswith('_'):
            self[name] = value
        else:
            super(ChainDict, self).__setattr__(name, value)

    def __and__(self, mode: int) -> Any: # &
        if self.__objects__ is None:
            raise EnvironmentError('The copied object does not support the & operator.')
        mode_ways = {
            SM.text: self.__objects__.find_text,
            SM.attr_names: self.__objects__.get_attr_names,
            SM.tag: self.__objects__.get_tag,
            SM.ancestor: self.__objects__.get_ancestor,
            SM.descendants: self.__objects__.get_descendants,
            SM.siblings: self.__objects__.get_siblings,
            SM.children: self.__objects__.get_children,
            SM.parent: self.__objects__.get_parent,
        }
        if mode in mode_ways:
            return mode_ways[mode](self)
        else:
            raise TypeError('Output mode error.')

class ChainExpressionManager:

    SEARCH_ATTR_KEY: str = 'id'

    def __init__(self, manager, objects, nodes) -> None:
        self.manager = manager
        self.objects = objects

        self.index: int = 0
        self.nodes = nodes

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.nodes):
            self.index += 1
            return self.nodes[self.index-1]
        else:
            raise StopIteration

    def __getitem__(self, key):
        return self.nodes[key]

    def __len__(self):
        return len(self.nodes)

    def __truediv__(self, text: str): # /
        self.index = 0
        temp_nodes = self.manager.find_nodes_by_text(text)
        descendants = reduce(lambda x,y:x+y, [_node&SM.descendants for _node in self.nodes]) if self.nodes else []
        self.nodes = find_intersection([self.nodes+descendants, temp_nodes])
        return self

    def __mod__(self, cdata: str): # %
        self.index = 0
        temp_nodes = self.manager.find_nodes_by_cdata(cdata)
        descendants = reduce(lambda x,y:x+y, [_node&SM.descendants for _node in self.nodes]) if self.nodes else []
        self.nodes = find_intersection([self.nodes+descendants, temp_nodes])
        return self

    def __floordiv__(self, comment: str): # //
        self.index = 0
        temp_nodes = self.manager.find_nodes_by_comment(comment)
        descendants = reduce(lambda x,y:x+y, [_node&SM.descendants for _node in self.nodes]) if self.nodes else []
        self.nodes = find_intersection([self.nodes+descendants, temp_nodes])
        return self

    def __or__(self, tag: str): # |
        self.index = 0
        temp_nodes = self.manager.find_nodes_by_tag(tag)
        descendants = reduce(lambda x,y:x+y, [_node&SM.descendants for _node in self.nodes]) if self.nodes else []
        self.nodes = find_intersection([self.nodes+descendants, temp_nodes])
        return self

    def __matmul__(self, attrs: Union[dict, str]): # @
        self.index = 0
        if isinstance(attrs, dict):
            temp_nodes = self.manager.find_nodes_by_attrs(**attrs)
        elif isinstance(attrs, str):
            temp_nodes = self.manager.find_nodes_by_attrs(**{self.SEARCH_ATTR_KEY: attrs})
        else:
            raise TypeError('`attrs` must be dict or str.')
        descendants = reduce(lambda x,y:x+y, [_node&SM.descendants for _node in self.nodes]) if self.nodes else []
        self.nodes = find_intersection([self.nodes+descendants, temp_nodes])
        return self

    def __xor__(self, count: int): # ^
        if 1 == count:
            if len(self.nodes) > 1:
                raise IndexError('More than one result.')
            if 0 <= count-1 < len(self.nodes):
                return self.nodes[count-1]
            else:
                raise IndexError('No nodes found.')
        elif count > 1:
            return self.nodes[:count]
        else:
            raise IndexError('`count` must be greater than or equal to 1.')
